'''
Date: 2024-08-16 09:36:38
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-09-03 16:14:07
Description: LazyDock Pymol Plugin
'''

import os

os.environ['MBAPY_FAST_LOAD'] = 'True'
os.environ['MBAPY_AUTO_IMPORT_TORCH'] = 'False'

import sys

sys.path.append(os.path.dirname(__file__))

from main import GUILauncher
from pymol import cmd

from lazydock.pml.server import VServer


def __init__(self):
    try:
        from pymol.plugins import addmenuitemqt
        addmenuitemqt('LazyDock', GUILauncher)
        return
    except Exception as e:
        print(e)
    self.menuBar.addmenuitem('Plugin', 'command', 'lazydock',
                             label = 'LazyDock', command = lambda s=self : GUILauncher(s)) 
    


def start_lazydock_server(host: str = 'localhost', port: int = 8085, quiet: int = 1):
    print(f'Starting LazyDock server on {host}:{port}, quiet={quiet}')
    VServer(host, port, not bool(quiet))
    
    
cmd.extend('start_lazydock_server', start_lazydock_server)


'''
Date: 2024-10-05 16:58:05
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-10-08 10:26:58
Description: 
'''
from DuIvyTools.DuIvyTools.FileParser.xvgParser import XVG
from DuIvyTools.DuIvyTools.FileParser.xpmParser import XPM
from DuIvyTools.DuIvyTools.Visualizer.Visualizer_matplotlib import LineMatplotlib


'''
Date: 2024-09-30 19:28:57
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-10-09 21:06:26
Description: RRCS calculation in PyMOL, RRCS is from article "Common activation mechanism of class A GPCRs": https://github.com/elifesciences-publications/RRCS/blob/master/RRCS.py
'''
from typing import Dict, Tuple

import pandas as pd
from pymol import cmd


def _test_close(dict_coord: Dict[str, Dict[int, Tuple[float, float, float, float]]],
                ires: str, jres: str):
    for iatom in dict_coord[ires]:
        (ix, iy, iz, iocc) = dict_coord[ires][iatom]
        for jatom in dict_coord[jres]:                  
            (jx, jy, jz, jocc) = dict_coord[jres][jatom]
            dx = abs(ix-jx)
            dy = abs(iy-jy)
            dz = abs(iz-jz)
            if dx < 4.63 and dy < 4.63 and dz < 4.63:
                return True
    return False

def _calcu_score(dict_coord: Dict[str, Dict[int, Tuple[float, float, float, float]]],
                 atomnum2name: Dict[int, str], ires: str, jres: str,
                 check_name: bool, score_count: int):
    total_score = 0
    for iatom in dict_coord[ires]:
        if check_name and atomnum2name[iatom] in ['N', 'CA', 'C', 'O']:
            continue
        (ix, iy, iz, iocc) = dict_coord[ires][iatom]
        for jatom in dict_coord[jres]:
            if check_name and atomnum2name[jatom] in ['N', 'CA', 'C', 'O']:
                continue
            (jx, jy, jz, jocc) = dict_coord[jres][jatom]
            d2 = (ix-jx)**2 + (iy-jy)**2 + (iz-jz)**2
            if d2 >= 21.4369:  # 4.63*4.63 = 21.4369
                score = 0
            elif d2 <= 10.4329:  # 3.23*3.23 = 10.4329
                score = 1.0*iocc*jocc
            else:
                score = (1-(d2**0.5 - 3.23)/1.4)*iocc*jocc
            total_score += score
            score_count[0] += 1
    return total_score

def calcu_RRCS(model: str):
    dict_coord = {} # dict to store coordinates. dict_coord[res][atom] = (x, y, z, occupancy)
    cmd.iterate_state(1, model, 'dict_coord.setdefault(f"{chain}:{resi}:{resn}", {}).setdefault(index, (x, y, z, q))', space={'dict_coord': dict_coord})
    atomnum2name = {} # map atom number to atom name, in order to find N, CA, C, O
    cmd.iterate(model, 'atomnum2name[index] = name', space={'atomnum2name': atomnum2name})
    contact_score = {} # dict to store final results. contact_score[ires][jres] = contact_score.
    score_count = [0] # 66320
    # calcu RRCS score for each residue pair
    for ires in dict_coord:
        ires_num = int(ires.split(':')[1])
        contact_score[ires] = {}
        # find jres if it has any atom close to ires
        for jres in dict_coord:
            jres_num = int(jres.split(':')[1])
            contact_score[ires][jres] = 0
            # skip because alreadly calculated
            if jres_num <= ires_num:
                continue
            # skip if jres has no atom close to ires
            if not _test_close(dict_coord, ires, jres):
                continue
            # calculate RRCS score for ires and jres
            contact_score[ires][jres] = _calcu_score(dict_coord, atomnum2name, ires, jres,
                                                     check_name= abs(ires_num - jres_num) < 5, score_count=score_count)
    # convert dict to DataFrame
    contact_df = pd.DataFrame(data={k:list(v.values()) for k,v in contact_score.items()},
                              index=contact_score.keys(), columns=contact_score.keys())
    return contact_df

if __name__ == '__main__':
    cmd.reinitialize()
    cmd.load('data_tmp/pdb/RECEPTOR.pdb', 'receptor')
    from mbapy.base import TimeCosts
    @TimeCosts(3)
    def test_calcu(idx):
        calcu_RRCS('receptor')
    test_calcu()


'''
Date: 2024-08-19 10:41:56
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-08-31 20:10:11
Description: 
'''
from threading import Thread

from lazy_dlg import LazyDLG
from lazy_plot import LazyPlot
from lazy_pocket import LazyPocket
# from mbapy.web import TaskPool
from nicegui import app, ui
from pymol import cmd


class GUILauncher:
    def __init__(self, app = None):
        
        self._now_molecule = cmd.get_names_of_type('object:molecule') or []
        self._now_selection = cmd.get_names_of_type('selection') + ['sele']
        self.ui_update_func = []
        ui.timer(1, self.ui_update_content_from_pymol)
        
        self.lazy_pocket = LazyPocket(self)
        self.lazy_dlg = LazyDLG(self)
        self.lazy_plot = LazyPlot(self)
        
        self.build_gui()
        
        self.host = Thread(name = 'GUI-Host', target=ui.run,
                            kwargs = dict(title='LazyDock', host = 'localhost', port=8090, reload = False),
                            daemon=False)
        self.host.start()
        
    def ui_update_content_from_pymol(self, ):
        self._now_molecule = cmd.get_names_of_type('object:molecule')
        self._now_selection = cmd.get_names_of_type('selection') + ['sele']
        
        for fn in self.ui_update_func:
            fn()
        
    def build_gui(self, ):
        with ui.header(elevated=True).style('background-color: #3874c8'):
            ui.label('LazyDock | Pymol Plugin').classes('text-h4')
            ui.space()
            ui.button('Exit', on_click=app.shutdown, icon='power')
        with ui.splitter(value=10).classes('w-full h-full') as splitter:
            with splitter.before:
                with ui.tabs().props('vertical').classes('w-full') as tabs:
                    lazy_pocket_tab = ui.tab('Pocket')
                    lazy_dlg_tab = ui.tab('DLG')
                    lazy_plot_tab = ui.tab('Plot')
            with splitter.after:
                with ui.tab_panels(tabs, value=lazy_pocket_tab) \
                        .props('vertical').classes('w-full h-full'):
                    with ui.tab_panel(lazy_pocket_tab):
                        self.lazy_pocket.build_gui()
                    with ui.tab_panel(lazy_dlg_tab):
                        self.lazy_dlg.build_gui()
                    with ui.tab_panel(lazy_plot_tab):
                        self.lazy_plot.build_gui()

    
if __name__ in {"__main__", "__mp_main__"}:
    # dev code
    app = GUILauncher()


'''
Date: 2024-08-15 19:54:22
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-08-21 15:25:15
Description: print selected residue names and numbers as autodock flex receptor residue format
'''
import os
import tkinter as tk
import uuid
from typing import List

from nicegui import ui
from pymol import CmdException, api, cmd

from lazydock.pml.autodock_utils import MyFileDialog
from lazydock.utils import uuid4

NULL_CHAIN = "''"


def _get_res_info_from_sele(select: str, attrs: List[str] = None):
    """
    Parameters
    ----------
        - select : str, pymol selection
        - attrs : List[str], optional, attributes to get, by default ['model', 'chain', 'resn', 'resi']

    Returns
    -------
        - chains : dict
            - key: model_str:chain_str
            - value: dict
                - key: res_str:res_id_str
                - value: True
        - chains_data : dict
            - key: model_str
            - value: dict
                - key: chain_str
                - value: List[List[res_str, res_id]]
    """
    if attrs is None:
        attrs = ['model', 'chain', 'resn', 'resi']
    atoms = {k:[] for k in attrs}
    chains = {}
    chains_data = {}
    
    cmd.iterate(select, lambda atom: atoms.setdefault('model', []).append(atom.model))
    cmd.iterate(select, lambda atom: atoms.setdefault('chain', []).append(atom.chain))
    cmd.iterate(select, lambda atom: atoms.setdefault('resn', []).append(atom.resn))
    cmd.iterate(select, lambda atom: atoms.setdefault('resi', []).append(atom.resi))
    
    for m, c, r, i in zip(atoms['model'], atoms['chain'], atoms['resn'], atoms['resi']):
        chains.setdefault(f'{m}:{c}', {})[f'{r}{i}'] = True
        chains_data.setdefault(m, {c:[]})[c].append([r, i])
        
    return chains, chains_data
    

class LazyPocket:
    def __init__(self, app, _dev_mode: bool = False):
            
        self._app = app
        self._app.ui_update_func.append(self.ui_update_ui)
        
        self.ui_molecule = None
        self.ui_sele = None
        self.ui_radius = None
        
        self.sele = None
        self.sele_molecule = None
        self.sele_selection = None
        self.radius = None
        self.sele_chains = None
        
    def ui_update_ui(self):
        self.ui_molecule.set_options(self._app._now_molecule)
        self.ui_sele.set_options(self._app._now_selection)
        
    def build_flex_gui(self):
        # sele
        with ui.row().classes('w-full'):
            self.ui_molecule = ui.select(self._app._now_molecule,
                                          label = 'select a molecule').bind_value_to(self, 'sele_molecule').classes('w-1/5').props('use-chips')
            ui.label('OR')
            self.ui_sele = ui.select(self._app._now_selection,
                                     label = 'select a selection').bind_value_to(self, 'sele_selection').classes('w-1/5').props('use-chips')
        # radius
        self.ui_radius = ui.number(label = 'Radius (A)', min = 0, step=0.1, value = 0).bind_value_to(self, 'radius').classes('w-1/5')
        # buttons
        with ui.row().classes('w-full'):
            self.ui_print_butt = ui.button(text = 'print', on_click=self.print_sele_around_res)
            self.ui_save_butt = ui.button(text ='save rigid', on_click=self.save_rigid_receptor)
        # logs
        self.log = ui.log(max_lines=20).classes('w-full h-2/5')
        
    def build_gui(self):
        with ui.tabs().classes('w-full').props('align=left active-bg-color=blue') as tabs:
            self.ui_flex_tab = ui.tab('AutoDock Flex Residue Helper').props('no-caps')
        with ui.tab_panels(tabs, value=self.ui_flex_tab).classes('w-full'):
            with ui.tab_panel(self.ui_flex_tab):
                self.build_flex_gui()
        # return self
        return self
        
    def print_sele_around_res(self):
        """output example: CB1:D:PHE108_PHE177_HIS178_LEU193_PHE379_SER383"""
        self.sele = self.sele_molecule or self.sele_selection
        if self.radius:
            pocket = f'Pocket_{uuid4()}'
            if self.sele_molecule:
                cmd.select(pocket, f'model {self.sele_molecule}')
                cmd.select(pocket, f'byres {pocket} around {self.radius}')
            else:
                cmd.select(pocket, f'byres {self.sele} around {self.radius}')
            self.sele = pocket
        chains, self.sele_chains = _get_res_info_from_sele(self.sele)
        final_output = ",".join(f"{k}:{'_'.join(v.keys())}" for k,v in chains.items())
        log = f'Residue names and numbers as autodock flex receptor residue format:\n{final_output}'
        print(log)
        self.log.push(log)
        
    def save_rigid_receptor(self):
        if self.sele_chains is None:
            ui.notify('Please run "Print sele (around pocket) in flex residue format" first.')
            return None
        pdb_path = MyFileDialog(types = [('PDB file', '*.pdb')],
                                initialdir=os.getcwd()).get_save_file()
        if pdb_path is None:
            ui.notify('Please choose a file to save.')
            return None
        if not pdb_path.endswith('.pdb'):
            pdb_path += '.pdb'
        is_first = True
        for model in self.sele_chains:
            tmp_model_name = f'{model}_{str(uuid.uuid4())[:4]}'
            api.copy(tmp_model_name, model)
            for chain in self.sele_chains[model]:
                for _, resi in self.sele_chains[model][chain]:
                    tmp_sele_name = f'{tmp_model_name}_{chain}_{resi}'
                    try:
                        api.select(tmp_sele_name, f'model {tmp_model_name} and (chain {chain or NULL_CHAIN} and resi {resi})')
                    except CmdException:
                        # TODO: don't kown why, but copy again helps
                        api.copy(tmp_model_name, model)
                        api.select(tmp_sele_name, f'model {tmp_model_name} and (chain {chain or NULL_CHAIN} and resi {resi})')
                    cmd.remove(tmp_sele_name)
                    cmd.delete(tmp_sele_name)
            api.multisave(pdb_path, tmp_model_name, append = 0 if is_first else 1)
            is_first = False
        log = f'Rigid receptor saved to {pdb_path}.'
        print(log)
        self.log.push(log)
            
        
# dev mode
if __name__ in {"__main__", "__mp_main__"}:    
    cmd.reinitialize()
    cmd.load('data_tmp/pdb/LIGAND.pdb', 'ligand')
    cmd.load('data_tmp/pdb/RECEPTOR.pdb', 'receptor')
    
    from main import GUILauncher
    GUILauncher()
    


# copied and fixed from Autodock Plugin: http://www.pymolwiki.org/index.php/autodock_plugin


from __future__ import absolute_import, print_function

import re
import tkinter as tk
import tkinter.filedialog as tkFileDialog
from typing import Any, Callable, List

from mbapy.base import put_err
from mbapy.file import decode_bits_to_str, opts_file
from mbapy.game import BaseInfo

# atom-type, atom-number, atom-name, residue-name, chain-name, residue-number, x, y, z, occupancy, temperature-factor
# ATOM      1  CA  LYS     7     136.747 133.408 135.880 -0.06 +0.10 OA
PDB_PATTERN = r"(ATOM|HETATM) +(\d+) +(\w+) +(\w+) +(\w+)? +(\d+) +([\d\-\.]+) +([\d\-\.]+) +([\d\-\.]+) +([\+\-][\d\-\.]+) +([\+\-][\d\-\.]+) [ \.\-\+\d]+ ([A-Z]+)"
PDB_FORMAT = "{:6s}{:>5s}  {:<3s} {:>3s} {:1s}{:>4s}    {:>8s}{:>8s}{:>8s}{:>6s}{:>6s}          {:>2s}  "
PDB_FORMAT2= "{:6s}{:>5s} {:<4s} {:>3s} {:1s}{:>4s}    {:>8s}{:>8s}{:>8s}{:>6s}{:>6s}          {:>2s}  "

class ADModel(BaseInfo):
    """STORAGE CLASS FOR DOCKED LIGAND"""
    def __init__(self, content: str = None, _sort_atom_by_res: bool = False,
                 _parse2std: bool = False):
        self.energy = 0.
        self.name = ''
        self.poseN = 0
        self.info = ''
        self.pdb_string = ''
        self.pdb_lines = []
        if content is not None:
            self.parse_content(content, _sort_atom_by_res, _parse2std)

    def parse_content(self, content: str, _sort_atom_by_res: bool = False,
                      _parse2std: bool = False):
        # parse pdb lines
        self.info = content
        self.pdb_lines = list(map(lambda x: x[0], re.findall(r'((ATOM|HETATM).+?\n)', self.info)))
        self.pdb_atoms = list(map(list, re.findall(PDB_PATTERN, self.info)))
        if len(self.pdb_atoms) != len(self.pdb_lines):
            put_err(f'pdb_atoms and pdb_lines length not match: {len(self.pdb_atoms)} vs {len(self.pdb_lines)}, just use pdb_lines')
            _sort_atom_by_res = _parse2std = False
        if _sort_atom_by_res:
            pack = sorted(zip(self.pdb_lines, self.pdb_atoms), key = lambda x : (x[1][4], int(x[1][5]), int(x[1][1])))
            self.pdb_lines, self.pdb_atoms = zip(*pack)
        if _parse2std:
            for atom in self.pdb_atoms:
                if not atom[4]:
                    atom[4] = 'A'
                if len(atom[2]) == 4:
                    atom[-1] = atom[2][1] # such as 1HH1
                elif len(atom[2]) >= 1:
                    if atom[2][0].isdigit():
                        atom[-1] = atom[2][1] # such as 1HH
                    else:
                        atom[-1] = atom[2][0] # such as H, CA, NH1
            self.pdb_lines = [PDB_FORMAT2.format(*line) if len(line[2])==4 else PDB_FORMAT.format(*line) for line in self.pdb_atoms]
            self.pdb_string = '\n'.join(self.pdb_lines)
        else:
            self.pdb_string = ''.join(self.pdb_lines)
        # parse energy could be str.find?
        energy_line = re.findall(r'USER.+?Free Energy of Binding.+?\n', self.info)
        if energy_line:
            entr = energy_line[0].split('=')[1]
            self.energy = float(entr.split()[0])
        else:
            energy_line = re.findall(r'REMARK.+?VINA RESULT.+?\n', self.info)
            if energy_line:
                entr = energy_line[0].split(':')[1]
                self.energy = float(entr.split()[0])
            else:
                self.energy = None

    def as_pdb_string(self):
        return self.pdb_string

    def info_string(self):
        return self.info


class DlgFile(BaseInfo):
    def __init__(self, path: str = None, content: str = None,
                 sort_pdb_line_by_res: bool = False,
                 parse2std: bool = False):
        # load content from file if path is provided
        self.path = path
        self.sort_pdb_line_by_res = sort_pdb_line_by_res
        self.parse2std = parse2std
        if path is not None:
            self.content = decode_bits_to_str(opts_file(path, 'rb'))
        elif content is not None:
            self.content = content
        # decode content to pose_lst
        if self.content is not None:
            self.pose_lst: List[ADModel] = self.decode_content()
            self.sort_pose()
        else:
            self.pose_lst: List[ADModel] = []
        self.n2i = {}
        
    def __len__(self):
        return len(self.pose_lst)
        
    def sort_pose(self, key: Callable[[ADModel], Any] = None,
                  inplace: bool = True, reverse: bool = False) -> None:
        if key is None:
            key = lambda x : x.energy
        if inplace:
            self.pose_lst.sort(key=key, reverse=reverse)
            ret = self.pose_lst
        else:
            ret = sorted(self.pose_lst, key=key, reverse=reverse)
        return ret
        
    def decode_content(self):
        dlg_pose_lst = []
        for model in re.findall('MODEL.+?ENDMDL', self.content, re.DOTALL):
            model = model.replace('\nDOCKED: ', '\n')
            dlg_pose_lst.append(ADModel(model, self.sort_pdb_line_by_res, self.parse2std))
        return dlg_pose_lst
    
    def asign_pose_name(self, pose_names: List[str]):
        if len(pose_names) != len(self.pose_lst):
            raise ValueError("Number of pose names must match number of poses")
        for i, name in enumerate(pose_names):
            self.n2i[name] = i
        
    def asign_prop(self, prop: str, value: List[Any]):
        if value is not None and len(value) == len(self.pose_lst):
            setattr(self, prop, value)
                
    def set_pose_prop(self, prop: str, value: Any, pose_name: str = None, pose_idx: int = None):
        if pose_name is None and pose_idx is None:
            raise ValueError("Either pose_name or pose_idx must be provided")
        if pose_name is not None:
            pose_idx = self.n2i[pose_name]
        if not hasattr(self, prop):
            setattr(self, prop, [None]*len(self.pose_lst))
        getattr(self, prop)[pose_idx] = value
            
    def get_pose(self, pose_name: str = None, pose_idx: int = None):
        if pose_name is None and pose_idx is None:
            raise ValueError("Either pose_name or pose_idx must be provided")
        if pose_name is not None:
            pose_idx = self.n2i[pose_name]
        return self.pose_lst[pose_idx]
    
    def get_pose_prop(self, prop: str, pose_name: str = None, pose_idx: int = None, default: Any = None):
        if pose_name is None and pose_idx is None:
            raise ValueError("Either pose_name or pose_idx must be provided")
        if pose_name is not None:
            pose_idx = self.n2i[pose_name]
        if hasattr(self, prop):
            return getattr(self, prop)[pose_idx]
        else:
            return default



def tk_file_dialog_wrapper(*args, **kwargs):
    def ret_wrapper(tk_file_dialog_func):
        def core_wrapper(*args, **kwargs):
            parent = tk.Tk()
            result = tk_file_dialog_func(*args, parent = parent, **kwargs)
            parent.withdraw()
            if result == "":
                return None
            else:
                return result
        return core_wrapper
    return ret_wrapper
            

class MyFileDialog:
    def __init__(self, types=[("Executable", "*")], initialdir: str = None):
        self.initialdir = initialdir
        self.types = types

    @tk_file_dialog_wrapper()
    def get_open_file(self, parent):
        return tkFileDialog.askopenfilename(parent=parent, initialdir = self.initialdir, filetypes=self.types)


    @tk_file_dialog_wrapper()
    def get_save_file(self, parent):
        return tkFileDialog.asksaveasfilename(parent=parent, initialdir = self.initialdir, filetypes=self.types)
        
    @tk_file_dialog_wrapper()
    def get_ask_dir(self, parent):
        return tkFileDialog.askdirectory(parent=parent, initialdir = self.initialdir)


if __name__ == '__main__':
    # from mbapy.base import TimeCosts
    # @TimeCosts(10)
    # def load_test(idx):
    #     DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=True)
    # load_test()
    normal = DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=False, parse2std=False)
    std = DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=False, parse2std=True)
    from pymol import cmd
    from rdkit import Chem
    cmd.reinitialize()
    cmd.read_pdbstr(std.pose_lst[0].as_pdb_string(), 'std')
    print(std.pose_lst[0].as_pdb_string(), '\n\n')
    print(cmd.get_pdbstr('std'))
    ligand = Chem.MolFromPDBBlock(cmd.get_pdbstr('std'), removeHs=True, sanitize=False)
    pass


'''
Date: 2024-08-18 12:56:06
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-08-31 21:49:34
Description: 
'''
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from mbapy.base import put_err
from pymol import CmdException, cmd

if __name__ == '__main__':
    from lazydock.utils import uuid4
else:
    from ..utils import uuid4

RES_SEPRATOR = '::'
NULL_CHAIN = "''"


def get_h_bond_cutoff():
    """ Get the distance cutoff (center and edge) for hydrogen bonds."""
     # ideal geometry and minimally acceptable geometry
    return cmd.get('h_bond_cutoff_center'), cmd.get('h_bond_cutoff_edge')


def set_h_bond_cutoff(center: float, edge: float):
    """ Set the distance cutoff (center and edge) for hydrogen bonds."""
    cmd.set('h_bond_cutoff_center', center)
    cmd.set('h_bond_cutoff_edge', edge)
    

def get_distance_info(dist_name: str, state = 1, selection = 'all', xyz2atom = None):
    '''
    Params:
        - dist_name : str, name of distance object
        - state : int, state of distance object
        - selection : str, selection of atoms to compute distances for
        - xyz2atom : dict, mapping of xyz coordinates to atom info: {(x,y,z): (model, chain, resn, resi, elem, index)}
        
    Returns:
        - list of tuples, each tuple contains two atom info and distance value, in the format of ((model, chain, resn, resi, elem, index), (model, chain, resn, resi, elem, index), distance)
    
    Notes:
        - the order of two sele in the return is not guaranteed to be the same as the order in the input selection
        
    modified from http://pymolwiki.org/index.php/get_raw_distances
    '''
    from chempy import cpv

    state = int(state)
    if state < 1:
        state = cmd.get_state()

    # Possible return values are

    # "object:molecule"
    # "object:map"
    # "object:mesh"
    # "object:slice"
    # "object:surface"
    # "object:measurement"
    # "object:cgo"
    # "object:group"
    # "object:volume"
    # "selection"
    if dist_name not in cmd.get_names_of_type('object:measurement'):
        return put_err('no such distance object: ' + dist_name)

    raw_objects = cmd.get_session(dist_name, 1, 1, 0, 0)['names']

    if xyz2atom is None:
        xyz2atom = {}
        cmd.iterate_state(state, selection, 'xyz2idx[x,y,z] = (model,chain,resn,resi,elem,index)',
                        space=locals())

    r = []
    for obj in raw_objects:
        try:
            points = obj[5][2][state - 1][1]
            if points is None:
                raise ValueError
        except (KeyError, ValueError):
            continue
        for i in range(0, len(points), 6):
            xyz1 = tuple(points[i:i + 3])
            xyz2 = tuple(points[i + 3:i + 6])
            try:
                r.append([xyz2atom[xyz1], xyz2atom[xyz2], cpv.distance(xyz1, xyz2)])
            except Exception as e:
                put_err(f'Error processing distance object {obj[0]}: {e}')
    return r


def calcu_atom_level_interactions(sele1: str, sele2: str, mode = 0, cutoff=3.6, xyz2atom: Dict = None):
    """
    calcu distances between atoms set sele1 and sele2 using cmd.distance, and get the distances and atom info.
    
    Parameters:
        - sele1: str, selection 1 of atoms
        - sele2: str, selection 2 of atoms
        - mode: int, mode of cmd.distance
        - cutoff: float, cutoff of cmd.distance
        - xyz2atom: dict, mapping of xyz coordinates to atom info: {(x,y,z): (model, chain, resn, resi, elem, index)}
        
    Returns:
        - atoms: dict, atom info for each selection, in the format of 
            {'s1_a': [(model, chain, resn, resi, elem, index),...],
             's1_d': [(model, chain, resn, resi, elem, index),...],
             's2_a': [(model, chain, resn, resi, elem, index),...],
             's2_d': [(model, chain, resn, resi, elem, index),...]}
        - xyz2atom: dict, mapping of xyz coordinates to atom info: {(x,y,z): (model, chain, resn, resi, elem, index)}
        - residues: dict, residue info for each selection, in the format of {'s1_a': {model: {chain: {resn-resi: elem}}...},'s1_d': {model: {chain: {resn-resi: elem}}...},'s2_a': {model: {chain: {resn-resi: elem}}...},'s2_d': {model: {chain: {resn-resi: elem}}...}}
        - interactions: dict, interactions between atoms in each selection, in the format of {'aa': [(('receptor', 'A', 'LYS', '108', 'O', 459), ('ligand', '', 'TYR', '1', 'N', 48), 3.4605595828662383),...], 'ad': [(('receptor', 'A', 'LYS', '108', 'O', 459), ('ligand', '', 'TYR', '1', 'N', 48), 3.4605595828662383),...], 'da': [(('receptor', 'A', 'LYS', '108', 'O', 459), ('ligand', '', 'TYR', '1', 'N',
    """
    interactions = {}
    rand_id = uuid4()
    atom_selections = {
        's1_a': f'{sele1}_a_{rand_id}',
        's1_d': f'{sele1}_d_{rand_id}',
        's2_a': f'{sele2}_a_{rand_id}',
        's2_d': f'{sele2}_d_{rand_id}',
    }
    dist_names = {
        'aa': f'aa_{rand_id}',
        'ad': f'ad_{rand_id}',
        'da': f'da_{rand_id}',
        'dd': f'dd_{rand_id}',
    }
    atoms = {K:[] for K in atom_selections.keys()}
    residues = {K:{} for K in atom_selections.keys()} # model:chain:resn-resi
    if xyz2atom is None:
        xyz2atom = {}
        cmd.iterate_state(1, 'all', 'xyz2atom[x,y,z] = (model,chain,resn,resi,elem,index)', space = locals())
        
    # make selections for each atom type in sele1 and sele2
    for atom_sele in atom_selections.values():
        sele = sele1 if atom_sele.startswith(sele1) else sele2
        atom_type1 = 'donor' if atom_sele == f'{sele}_d_{rand_id}' else 'acceptor'
        atom_type2 = 'acceptor' if atom_sele == f'{sele}_d_{rand_id}' else 'donor'
        cmd.select(atom_sele, f'{sele} and ({atom_type1} and !{atom_type2})')
    # get atom info in each selection
    for atom_type in atom_selections:
        cmd.iterate(atom_selections[atom_type], lambda atom: atoms[atom_type].append(atom))
    # ganther atoms to residues
    for atom_type in atom_selections:
        for atom in atoms[atom_type]:
            # if chain is null string, set it to ''
            residues[atom_type].setdefault(RES_SEPRATOR.join([atom.model, atom.chain or NULL_CHAIN, atom.resn, atom.resi]), atom.elem)
    
    # compute contacts between each pair of atoms in each selection
    for s1_atoms, s2_atoms, dist_name in [(atom_selections['s1_a'], atom_selections['s2_d'], dist_names['ad']),
                            (atom_selections['s1_d'], atom_selections['s2_a'], dist_names['da']),
                            (atom_selections['s1_a'], atom_selections['s2_a'], dist_names['aa']),
                            (atom_selections['s1_d'], atom_selections['s2_d'], dist_names['dd'])]:
        cmd.distance(dist_name, s1_atoms, s2_atoms, cutoff=cutoff, mode=mode)
        # get distances and build interactions
        interactions[dist_name[:2]] = get_distance_info(dist_name, xyz2atom=xyz2atom)
        
    # delete atom selctions
    for atom_sele in atom_selections.values():
        cmd.delete(atom_sele)
    # delete distance objects
    for dist_name in dist_names.values():
        cmd.delete(dist_name)
    
    return atoms, xyz2atom, residues, interactions


def sort_atom_level_interactions(interactions: Dict[str, List[Tuple[Tuple[str, str, str, str, str, float],
                                                         Tuple[str, str, str, str, str, float], float]]],
                                model1: str, model2: str):
    """sort the interactions returned by calcu_atom_level_interactions, sort the order in tuple by model1 and model2."""
    for ty in list(interactions.keys()):
        values = interactions[ty]
        for i in range(len(values)):
            if values[i][0][0] == model2:
                interactions[ty][i][0], interactions[ty][i][1] = interactions[ty][i][1], interactions[ty][i][0]
            elif values[i][0][0] != model1:
                return put_err(f'{values[i][0][0]} is not in {model1} and {model2}, abort sort', interactions)
    return interactions



def merge_interaction_df(interaction: Dict[str, List[Tuple[Tuple[str, str, str, str, str, float],
                                                            Tuple[str, str, str, str, str, float], float]]],
                         interaction_df: pd.DataFrame,
                         distance_cutoff: float, nagetive_factor: float):
    """merge the interactions returned by calcu_atom_level_interactions to interaction_df."""
    # index format: CHAIN_ID:RESI:RESN
    def set_points(ty: str, points: float, nagetive_factor: float):
        points = distance_cutoff - points
        if ty in {'ad', 'da'}:
            return points
        return nagetive_factor * points
    for interaction_type, values in interaction.items():
        for single_inter in values:
            # single_inter: (('receptor', 'A', 'LYS', '108', 'O', 459), ('ligand', '', 'TYR', '1', 'N', 48), 3.4605595828662383)
            receptor_res = f'{single_inter[0][1]}:{single_inter[0][3]}:{single_inter[0][2]}'
            ligand_res = f'{single_inter[1][1]}:{single_inter[1][3]}:{single_inter[1][2]}'
            points = set_points(interaction_type, single_inter[2], nagetive_factor)
            if ligand_res not in interaction_df.index or receptor_res not in interaction_df.columns:
                interaction_df.loc[ligand_res, receptor_res] = points
            elif np.isnan(interaction_df.loc[ligand_res, receptor_res]):
                interaction_df.loc[ligand_res, receptor_res] = points
            else:
                interaction_df.loc[ligand_res, receptor_res] += points
    return interaction_df


def calcu_receptor_poses_interaction(receptor: str, poses: List[str], mode: int = 0,
                                     cutoff: float = 4., nagetive_factor: float = -1.):
    """
    calcu interactions between one receptor and one ligand with many poses.
    
    Parameters:
        receptor: str, receptor pymol name
        poses: list of str, ligand pymol names
        mode: int, mode of cmd.distance
        cutoff: float, cutoff of cmd.distance
        nagetive_factor: float, factor to multiply the distance value for interations between acceptor and acceptor, and donor and donor.
        
    Returns:
        interactions (dict): interactions between receptor and ligand, in the format of {'ligand': [xyz2atom, residues, interactions]}, where xyz2atom is a dict, residues is a dict, interactions is a dict.
        
        interaction_df (pd.DataFrame): , interactions between receptor and ligand, in the format of ligand-residue-residue matrix, with the value of each cell is the interaction score between two atoms.
            interaction_df.loc[ligand_res, receptor_res] = score
    """
    def sort_func(index: pd.Index):
        index = index.str.split(':')
        return list(map(lambda x: (x[0], int(x[1]), x[2]), index))
    # prepare interactions
    all_interactions, interaction_df = {}, pd.DataFrame()
    # select receptor
    sele_receptor = uuid4()
    cmd.select(sele_receptor, receptor)
    # calcu for each ligand
    for ligand in poses:
        # calcu interaction
        sele_ligand = uuid4()
        cmd.select(sele_ligand, ligand)
        _, xyz2atom, residues, interactions = calcu_atom_level_interactions(sele_receptor, sele_ligand,
                                                                            mode, cutoff)
        interactions = sort_atom_level_interactions(interactions, receptor, ligand)
        # NOTE: do not save atoms, because pymol.editing._AtomProxy class can't be pickled
        all_interactions[ligand] = [xyz2atom, residues, interactions]
        cmd.delete(sele_ligand)
        # merge interactions by res
        merge_interaction_df(interactions, interaction_df, cutoff, nagetive_factor)
    cmd.delete(sele_receptor)
    if not interaction_df.empty:
        # sort res
        interaction_df.sort_index(axis=0, inplace=True, key=sort_func)
        interaction_df.sort_index(axis=1, inplace=True, key=sort_func)
        interaction_df.fillna(0, inplace=True)
    else:
        return None, None
    return interactions, interaction_df


def filter_interaction_df(interaction_df: pd.DataFrame, colum_axis_min: float = None,
                          row_axis_min: float = None, inplace: bool = False):
    """
    filter the interaction_df by the minimum value of each axis.
    
    Parameters:
        interaction_df(pd.DataFrame) : interactions between receptor and ligand, in the format of ligand-residue-residue matrix, with the value of each cell is the distance between two atoms.
        colum_axis_min(float) : minimum value of each column axis, None means no filter.
        row_axis_min(float) : minimum value of each row axis, None means no filter.
        inplace(bool) : , whether to filter the original interaction_df or return a deep copy.
        
    Returns:
        interaction_df(pd.DataFrame) : filtered interactions between receptor and ligand,
                    in the format of ligand-residue-residue matrix,
                    with the value of each cell is the distance between two atoms.
    """
    if inplace:
        tmp_interaction_df = interaction_df
    else:
        tmp_interaction_df = interaction_df.copy(deep = True)
    # filter ligand
    if row_axis_min is not None:
        for ligand_res in tmp_interaction_df.index:
            if tmp_interaction_df.loc[ligand_res].abs().max() < row_axis_min:
                tmp_interaction_df.drop(ligand_res, inplace=True)
    # filter receptor
    if colum_axis_min is not None:
        for receptor_res in tmp_interaction_df.columns:
            if tmp_interaction_df[receptor_res].abs().max() < colum_axis_min:
                tmp_interaction_df.drop(receptor_res, axis=1, inplace=True)
    return tmp_interaction_df


__all__ = [
    'get_h_bond_cutoff',
    'set_h_bond_cutoff',
    'get_distance_info',
    'calcu_atom_level_interactions',
    'sort_atom_level_interactions',
    'merge_interaction_df',
    'calcu_receptor_poses_interaction',
    'filter_interaction_df',
    ]


if __name__ == '__main__':
    # dev code
    cmd.reinitialize()
    cmd.load('data_tmp/pdb/RECEPTOR.pdb', 'RECEPTOR')
    cmd.load('data_tmp/pdb/LIGAND.pdb', 'LIGAND')
    cmd.select('sele1', 'RECEPTOR')
    cmd.select('sele2', 'LIGAND')
    atoms, xyz2atom, residues, interactions = calcu_atom_level_interactions('sele1', 'sele2')
    calcu_receptor_poses_interaction('RECEPTOR', ['LIGAND'])